<?php

namespace Pelecardwc;
use Pelecardwc\Traits\Singleton;
class Transaction extends \WC_Data {

	const META_KEY = '_wppc_transaction';
	const ACTION_TYPE_J2 = '2';
	const ACTION_TYPE_J4 = '4';
	const ACTION_TYPE_J5 = '5';

	protected $id = '';

	protected $data = [
		'status_code' => '',
		'error_message' => '',
	];

	protected $validate = true;

	public function __construct( string $transaction_id = '' ) {
		parent::__construct();

		$this->set_id( $transaction_id );

		if ( ! empty( $this->get_id() ) ) {
			$this->set_data( $this->get_transaction() );
		}
	}

	public function wppc_set_order_id( int $order_id ) {
		$this->add_meta_data( 'ParamX', $order_id, true );

		return $this;
	}

	public function get_id(): string {
		return $this->id;
	}

	public function set_id( $transaction_id ) {
		$this->id = $transaction_id;

		return $this;
	}

	public function set_data( array $data ) {
		$this->set_props( [
			'status_code' => $data['StatusCode'] ?? '',
			'error_message' => $data['ErrorMessage'] ?? '',
		] );

		if ( ! empty( $data['ResultData'] ) ) {
			foreach ( $data['ResultData'] as $key => $value ) {
				$this->add_meta_data( $key, $value, true );
			}
		}

		if ( ! empty( $data['UserData'] ) ) {
			foreach ( $data['UserData'] as $key => $value ) {
				$this->add_meta_data( $key, $value, true );
			}
		}

		if ( ! empty( $data['EZCountData'] ) ) {
			foreach ( $data['EZCountData'] as $key => $value ) {
				$this->add_meta_data( $key, $value, true );
			}
		}

		if ( ! empty( $data['ICountData'] ) ) {
			foreach ( $data['ICountData'] as $key => $value ) {
				$this->add_meta_data( $key, $value, true );
			}
		}

		if ( ! empty( $data['TamalData'] ) ) {
			foreach ( $data['TamalData'] as $key => $value ) {
				$this->add_meta_data( $key, $value, true );
			}
		}

		if ( empty( $this->get_id() ) ) {
			$this->set_id( $this->get_meta( 'PelecardTransactionId' ) );
		}

		$this->set_object_read( true );

		return $this;
	}

	public function get_transaction(): array {
		$result = $this->is_uuid() ? get_transaction_by_id( $this->get_id() ) : [];

		return is_wp_error( $result ) ? [] : $result;
	}

	public function is_uuid(): bool {
		$pattern = '/^[0-9A-F]{8}-[0-9A-F]{4}-4[0-9A-F]{3}-[89AB][0-9A-F]{3}-[0-9A-F]{12}$/i';

		return (bool) preg_match( $pattern, $this->get_id() );
	}

	public function get_order_id( $context = 'view' ) {
		if ( $this->meta_exists( 'AdditionalDetailsParamX' ) ) {
			$order_id = $this->get_meta( 'AdditionalDetailsParamX', true, $context );
		} elseif ( $this->meta_exists( 'ParamX' ) ) {
			$order_id = $this->get_meta( 'ParamX', true, $context );
		} else {
			$order_id = 0;
		}

		return (int) apply_filters( 'wppc/transaction/order_id', $order_id, $context, $this );
	}

	public function set_json_data( $data ) {
		if ( is_string( $data ) ) {
			$data = json_decode( $data, true );
		}

		$this->set_props( $data );

		if ( ! empty( $data['meta_data'] ) ) {
			$this->set_meta_data( $data['meta_data'] );
		}

		return $this;
	}

	public function set_meta_data( $data ) {
		if ( empty( $data ) || ! is_array( $data ) ) {
			return;
		}

		$this->maybe_read_meta_data();

		foreach ( $data as $meta ) {
			$meta = (array) $meta;
			if ( isset( $meta['key'], $meta['value'] ) ) {
				$this->meta_data[] = new \WC_Meta_Data( [
					'key' => $meta['key'],
					'value' => $meta['value'],
				] );
			}
		}
	}

	public function set_validate( bool $validate ) {
		$this->validate = $validate;

		return $this;
	}

	public function needs_validation(): bool {
		return apply_filters( 'wppc/transaction/needs_validation', $this->validate, $this );
	}

	public function set_status_code( string $status_code ) {
		$this->set_prop( 'status_code', $status_code );

		return $this;
	}

	public function set_error_message( string $error_message ) {
		$this->set_prop( 'error_message', $error_message );

		return $this;
	}

	public function get_error_message( $context = 'view' ) {
		return $this->get_prop( 'error_message', $context );
	}

	public function validate(): bool {
		$confirmation_key = $this->get_meta( 'ConfirmationKey' );
		$debit_total = $this->get_meta( 'DebitTotal' );
		$unique_key = $this->wppc_get_unique_key();

		return validate_by_unique_key( $confirmation_key, $debit_total, $unique_key );
	}

	public function wppc_get_unique_key() {
		$order = $this->get_order();
		if ( $order ) {
			return $order->get_order_key();
		}

		return Gateway::instance()->wppc_get_user_nonce( $this->get_user_id() );
	}

	public function get_order() {
		$order_id = $this->get_order_id();

		return wc_get_order( $order_id );
	}

	public function save() {
		$order = $this->get_order();
		if ( ! $order || $this->exists( $order ) ) {
			return;
		}

		$order->add_meta_data( self::META_KEY, (string) $this );
		$order->save();

		do_action( 'wppc/transaction/after_save', $this, $order );
	}

	public function is_token_valid(): bool {
		return 0 < $this->get_user_id() && ! empty( $this->get_token() );
	}

	public function get_user_id( $context = 'view' ): int {
		return (int) $this->get_meta( 'UserData1', true, $context );
	}

	public function wppc_get_timeout_redirect_url( $context = 'view' ): string {
		return $this->get_meta( 'UserData2', true, $context );
	}

	public function exists( \WC_Order $order ) {
		$transactions = Order::instance()->wppc_get_transactions( $order );

		return array_reduce( $transactions, function( $carry, $transaction ) {
			return $carry || $transaction->get_id() === $this->get_id();
		}, false );
	}

	public function flatten(): array {
		$flatten = [
			'StatusCode' => $this->get_status_code(),
			'ErrorMessage' => $this->get_error_message(),
		];

		foreach ( $this->get_meta_data() as $meta ) {
			$flatten[ $meta->get_data()['key'] ] = $meta->get_data()['value'];
		}

		return $flatten;
	}

	public function get_token( $context = 'view' ): string {
		return $this->get_meta( 'Token', true, $context );
	}

	public function wppc_debit_approve_number( $context = 'view' ): string {
		return $this->get_meta( 'DebitApproveNumber', true, $context );
	}

	public function wppc_get_token_object( Gateway $gateway ): \WC_Payment_Token_CC {
		$token = new \WC_Payment_Token_CC();
		$token->set_gateway_id( $gateway->id );
		$token->set_token( $this->get_token() );
		$token->set_last4( $this->wppc_get_last4() );
		$token->set_card_type( $this->wppc_get_card_type() );
		$token->set_expiry_year( $this->wppc_get_expiry_year() );
		$token->set_expiry_month( $this->wppc_get_expiry_month() );
		$token->set_user_id( $this->get_user_id() );

		return $token;
	}

	public function wppc_get_last4() {
		return substr( $this->wppc_get_card_number(), -4 );
	}

	public function wppc_get_card_number( $context = 'view' ): string {
		return $this->get_meta( 'CreditCardNumber', true, $context );
	}

	public function wppc_get_card_type(): string {
		$brand = $this->wppc_get_card_brand();
		$company = $this->wppc_get_card_company();

		return ( 0 < $brand )
			? $this->wppc_get_card_type_by_brand( $brand )
			: $this->wppc_get_card_type_by_company( $company );
	}

	public function wppc_get_card_brand( $context = 'view' ): int {
		return (int) $this->get_meta( 'CreditCardBrand', true, $context );
	}

	public function wppc_get_card_company( $context = 'view' ): int {
		return (int) $this->get_meta( 'CreditCardCompanyClearer', true, $context );
	}

	public function wppc_get_card_type_by_brand( $brand ) {
		$brands = apply_filters( 'wppc/card_brands', [
			1 => 'mastercard',
			2 => 'visa',
			3 => 'maestro',
			5 => 'isracard',
		] );

		return $brands[ $brand ] ?? $brand;
	}

	public function wppc_get_card_type_by_company( $company ) {
		$companies = apply_filters( 'wppc/card_companies', [
			1 => 'isracard',
			2 => 'visa',
			3 => 'diners',
			4 => 'american express',
			6 => 'leumi card',
		] );

		return $companies[ $company ] ?? $company;
	}

	public function wppc_get_expiry_year() {
		$expiry_year = substr( $this->wppc_get_card_expiry(), -2 );

		return date_create_from_format( 'y', $expiry_year )->format( 'Y' );
	}

	public function wppc_get_card_expiry( $context = 'view' ): string {
		return $this->get_meta( 'CreditCardExpDate', true, $context );
	}

	public function wppc_get_expiry_month() {
		return substr( $this->wppc_get_card_expiry(), 0, 2 );
	}

	public function wppc_get_total_payments( $context = 'view' ): int {
		$total_payments = (int) $this->get_meta( 'TotalPayments', true, $context );

		return $total_payments ?: 1;
	}

	public function is_success(): bool {
		return '000' === $this->get_status_code();
	}

	public function get_status_code( $context = 'view' ): string {
		return $this->get_prop( 'status_code', $context );
	}

	public function is_timeout(): bool {
		return 301 === (int) $this->get_status_code();
	}

	public function is_3ds_failure(): bool {
		return 650 === (int) $this->get_status_code();
	}

	public function get_3ds_params(): array {
		$params = array_filter( [
			$this->get_meta( 'Eci' ),
			$this->get_meta( 'XID' ),
			$this->get_meta( 'Cavv' ),
		] );

		return 3 === count( $params ) ? $params : [];
	}

	public function wppc_get_action_type( $context = 'view' ): string {
		return $this->get_meta( 'JParam', true, $context );
	}

	public function is_action_type( string $action_type ): bool {
		return $action_type === $this->wppc_get_action_type();
	}
}
